@@ -4,22 +4,23 @@ import shutil
 import tarfile
 import time
 import uuid
-from datetime import datetime
 from datetime import date
+from datetime import datetime
 from http.client import BadStatusLine
 from http.client import RemoteDisconnected
+from json import JSONEncoder
 from tempfile import gettempdir
 from tempfile import NamedTemporaryFile
 
 import pytest
 from ibutsu_client import ApiClient
 from ibutsu_client import ApiException
-from ibutsu_client import ArtifactApi
 from ibutsu_client import Configuration
-from ibutsu_client import HealthApi
-from ibutsu_client import ResultApi
-from ibutsu_client import RunApi
-from json import JSONEncoder
+from ibutsu_client.api.artifact_api import ArtifactApi
+from ibutsu_client.api.health_api import HealthApi
+from ibutsu_client.api.result_api import ResultApi
+from ibutsu_client.api.run_api import RunApi
+from ibutsu_client.exceptions import ApiValueError
 from urllib3.exceptions import MaxRetryError
 from urllib3.exceptions import ProtocolError
 
@@ -38,14 +39,14 @@ BLOCKER_CATEGORY_TO_CLASSIFICATION = {
 }
 
 # Place a limit on the file-size we can upload for artifacts
-UPLOAD_LIMIT = 256 * 1024  # 256 KiB
+UPLOAD_LIMIT = 5 * 1024 * 1024  # 5 MiB
 
 # Maximum number of times an API call is retried
 MAX_CALL_RETRIES = 3
 
 
 class DateTimeEncoder(JSONEncoder):
-    """ Handle datetime objects in the archiver. """
+    """Handle datetime objects in the archiver."""
 
     def default(self, obj):
         if isinstance(obj, (date, datetime)):
@@ -140,12 +141,14 @@ class IbutsuArchiver(object):
     frontend = None
 
     def __init__(self, source=None, path=None, extra_data=None):
-        self._results = {}
+        self.results = {}
         self._run_id = None
         self.run = None
         self._temp_path = path
         self.source = source or "local"
-        self.extra_data = extra_data or {}
+        self.extra_data = extra_data or {"component": None, "env": None}
+        # pytest session object, to be set by pytest_collection_modifyitems below
+        self._session = None
 
         # Set an env var ID
         if os.environ.get("IBUTSU_ENV_ID"):
@@ -219,12 +222,24 @@ class IbutsuArchiver(object):
 
     def shutdown(self):
         # Gather the summary before building the archive
-        summary = {"failures": 0, "skips": 0, "errors": 0, "xfailures": 0, "xpasses": 0, "tests": 0}
-        for result in self._results.values():
+        summary = {
+            "failures": 0,
+            "skips": 0,
+            "errors": 0,
+            "xfailures": 0,
+            "xpasses": 0,
+            "tests": 0,
+            "collected": 0,
+        }
+        for result in self.results.values():
             key = self._status_to_summary(result["result"])
             if key in summary:
                 summary[key] += 1
+            # update the number of tests that actually ran
             summary["tests"] += 1
+        # store the number of tests that were collected
+        summary["collected"] = getattr(self._session, "testscollected", summary["tests"])
+        # store the summary on the run
         self.run["summary"] = summary
         self.update_run()
         # Build the tarball
@@ -240,7 +255,8 @@ class IbutsuArchiver(object):
     def get_run_id(self):
         if not self.run:
             run = {
-                "duration": 0,
+                "duration": 0.0,
+                "component": "",
                 "summary": {
                     "failures": 0,
                     "skips": 0,
@@ -251,7 +267,7 @@ class IbutsuArchiver(object):
                 },
                 "metadata": self.extra_data,
                 "source": getattr(self, "source", "local"),
-                "start_time": datetime.utcnow(),
+                "start_time": datetime.utcnow().isoformat(),
             }
             self.run = self.add_run(run=run)
         return self.run["id"]
@@ -289,7 +305,7 @@ class IbutsuArchiver(object):
         result["metadata"].update(self.extra_data)
         with open(os.path.join(art_path, "result.json"), "w") as f:
             json.dump(result, f, cls=DateTimeEncoder)
-        self._results[result_id] = result
+        self.results[result_id] = result
         return result
 
     def update_result(self, id, result):
@@ -300,12 +316,18 @@ class IbutsuArchiver(object):
         result["metadata"].update(self.extra_data)
         with open(os.path.join(art_path, "result.json"), "w") as f:
             json.dump(result, f, cls=DateTimeEncoder)
-        self._results[id] = result
+        self.results[id] = result
 
-    def upload_artifact(self, id, filename, data):
+    def upload_artifact(self, id_, filename, data, is_run=False):
+        """
+        Save an artifact to the archive.
+
+        'id_' can be either a run or result ID.
+        'is_run' should be True if it is a run ID.
+        """
         file_size = os.stat(data).st_size
         if file_size < UPLOAD_LIMIT:
-            art_path = os.path.join(self.temp_path, id)
+            art_path = os.path.join(self.temp_path, id_)
             os.makedirs(art_path, exist_ok=True)
             shutil.copyfile(data, os.path.join(art_path, filename))
         else:
@@ -315,15 +337,15 @@ class IbutsuArchiver(object):
                 f" File will not be uploaded to Ibutsu."
             )
 
-    def upload_artifact_raw(self, res_id, filename, data):
+    def upload_artifact_raw(self, id_, filename, data, is_run=False):
         file_object = NamedTemporaryFile(delete=False)
         os_file_name = file_object.name
         file_object.write(data)
         file_object.close()
-        self.upload_artifact(res_id, filename, os_file_name)
+        self.upload_artifact(id_, filename, os_file_name, is_run=is_run)
 
-    def upload_artifact_from_file(self, res_id, logged_filename, filename):
-        self.upload_artifact(res_id, logged_filename, filename)
+    def upload_artifact_from_file(self, id_, logged_filename, filename, is_run=False):
+        self.upload_artifact(id_, logged_filename, filename, is_run=is_run)
 
     def get_xfail_reason(self, data, report):
         xfail_reason = None
@@ -357,7 +379,7 @@ class IbutsuArchiver(object):
         return skip_reason
 
     def get_classification(self, reason):
-        """ Get the skip/xfail classification and category from the reason"""
+        """Get the skip/xfail classification and category from the reason"""
         category = None
         try:
             category = reason.split("category:")[1].strip()
@@ -366,7 +388,11 @@ class IbutsuArchiver(object):
         return BLOCKER_CATEGORY_TO_CLASSIFICATION.get(category)
 
     @pytest.mark.tryfirst
-    def pytest_collection_modifyitems(self, items):
+    def pytest_collection_modifyitems(self, session, items):
+        # save the pytest session object for later use
+        self._session = session
+
+        # loop over all items and add ibutsu data
         for item in items:
             data = getattr(item, "_ibutsu", {})
             new_data = {"id": None, "data": {"metadata": {}}, "artifacts": {}}
@@ -382,7 +408,7 @@ class IbutsuArchiver(object):
                 params = {}
         else:
             params = {}
-        start_time = datetime.utcnow()
+        start_time = datetime.utcnow().isoformat()
         fspath = item.location[0] or item.fspath.strpath
         if "site-packages/" in fspath:
             fspath = fspath[fspath.find("site-packages/") + 14 :]
@@ -392,6 +418,7 @@ class IbutsuArchiver(object):
             "params": params,
             "start_time": start_time,
             "test_id": get_test_idents(item)[0],
+            "duration": 0.0,
             "metadata": {
                 "statuses": {},
                 "run": self.run_id,
@@ -420,6 +447,8 @@ class IbutsuArchiver(object):
         merge_dicts(old_data, data)
         item._ibutsu["data"] = data
         yield
+        # Finish up with the result and update it
+        self.update_result(item._ibutsu["id"], result=item._ibutsu["data"])
 
     def pytest_exception_interact(self, node, call, report):
         if not hasattr(report, "_ibutsu"):
@@ -434,7 +463,7 @@ class IbutsuArchiver(object):
         self.upload_artifact_raw(id, "traceback.log", bytes(report.longreprtext, "utf8"))
         data["metadata"]["short_tb"] = short_tb
         data["metadata"]["exception_name"] = call.excinfo.type.__name__
-        self.update_result(id, result=data)
+        report._ibutsu["data"] = data
 
     def pytest_runtest_logreport(self, report):
         if not hasattr(report, "_ibutsu"):
@@ -445,7 +474,6 @@ class IbutsuArchiver(object):
         else:
             xfail = False
 
-        id = report._ibutsu["id"]
         data = report._ibutsu["data"]
         data["metadata"]["user_properties"] = {key: value for key, value in report.user_properties}
         data["metadata"]["statuses"][report.when] = (report.outcome, xfail)
@@ -465,8 +493,8 @@ class IbutsuArchiver(object):
             classification = self.get_classification(reason)
             if classification:
                 data["metadata"]["classification"] = classification
-        data["duration"] = sum([v for v in data["metadata"]["durations"].values()])
-        self.update_result(id, result=data)
+        data["duration"] = sum(v for v in data["metadata"]["durations"].values())
+        report._ibutsu["data"] = data
 
     def pytest_sessionfinish(self):
         self.stop_timer()
@@ -484,13 +512,13 @@ class IbutsuSender(IbutsuArchiver):
     An enhanced Ibutsu plugin that also sends Ibutsu results to an Ibutsu server
     """
 
-    def __init__(self, server_url, source=None, path=None, extra_data=None):
+    def __init__(self, server_url, source=None, path=None, extra_data=None, token=None):
         self.server_url = server_url
         self._has_server_error = False
         self._server_error_tbs = []
         self._sender_cache = []
 
-        config = Configuration()
+        config = Configuration(access_token=token)
         config.host = self.server_url
         # Only set the SSL CA cert if one of the environment variables is set
         for env_var in CA_BUNDLE_ENVS:
@@ -518,7 +546,7 @@ class IbutsuSender(IbutsuArchiver):
                     return out
                 except (RemoteDisconnected, ProtocolError, BadStatusLine):
                     retries += 1
-            raise TooManyRetriesError('Too many retries while trying to call API')
+            raise TooManyRetriesError("Too many retries while trying to call API")
         except (MaxRetryError, ApiException, TooManyRetriesError) as e:
             self._has_server_error = self._has_server_error or True
             self._server_error_tbs.append(str(e))
@@ -538,7 +566,7 @@ class IbutsuSender(IbutsuArchiver):
         if server_run:
             self.run = server_run.to_dict()
 
-    def update_run(self, duration=None):
+    def update_run(self, duration=0.0):
         super().update_run(duration)
         self._make_call(self.run_api.update_run, self.run["id"], run=self.run)
 
@@ -555,11 +583,36 @@ class IbutsuSender(IbutsuArchiver):
         self._make_call(self.result_api.update_result, id, result=result, async_req=True)
         super().update_result(id, result)
 
-    def upload_artifact(self, id, filename, data):
-        super().upload_artifact(id, filename, data)
+    def upload_artifact(self, id_, filename, data, is_run=False):
+        super().upload_artifact(id_, filename, data, is_run=is_run)
         file_size = os.stat(data).st_size
         if file_size < UPLOAD_LIMIT:
-            self._make_call(self.artifact_api.upload_artifact, id, filename, data)
+            with open(data, "rb") as file_content:
+                try:
+                    if not file_content.closed:
+                        if is_run:
+                            # id_ is the run_id, we don't check the return_type because
+                            # artifact.to_dict() in the controller contains a None value
+                            self._make_call(
+                                self.artifact_api.upload_artifact,
+                                filename,
+                                file_content,
+                                run_id=id_,
+                                _check_return_type=False,
+                            )
+                        else:
+                            # id_ is the result_id, we don't check the return_type because
+                            # artifact.to_dict() in the controller contains a None valued
+                            self._make_call(
+                                self.artifact_api.upload_artifact,
+                                filename,
+                                file_content,
+                                result_id=id_,
+                                _check_return_type=False,
+                            )
+
+                except ApiValueError:
+                    print(f"Uploading artifact '{filename}' failed as the file closed prematurely.")
 
     def output_msg(self):
         with open(".last-ibutsu-run-id", "w") as f:
@@ -590,6 +643,7 @@ class IbutsuSender(IbutsuArchiver):
 
 def pytest_addoption(parser):
     parser.addini("ibutsu_server", help="The Ibutsu server to connect to")
+    parser.addini("ibutsu_token", help="The JWT token to authenticate with the server")
     parser.addini("ibutsu_source", help="The source of the test run")
     parser.addini("ibutsu_metadata", help="Extra metadata to include with the test results")
     parser.addini("ibutsu_project", help="Project ID or name")
@@ -602,6 +656,14 @@ def pytest_addoption(parser):
         default=None,
         help="URL for the Ibutsu server",
     )
+    group.addoption(
+        "--ibutsu-token",
+        dest="ibutsu_token",
+        action="store",
+        metavar="TOKEN",
+        default=None,
+        help="The JWT token to authenticate with the server",
+    )
     group.addoption(
         "--ibutsu-source",
         dest="ibutsu_source",
@@ -642,6 +704,9 @@ def pytest_configure(config):
         ibutsu_server = config.getini("ibutsu_server")
     if not ibutsu_server:
         return
+    ibutsu_token = config.getoption("ibutsu_token", None)
+    if config.getini("ibutsu_token"):
+        ibutsu_token = config.getini("ibutsu_token")
     ibutsu_source = config.getoption("ibutsu_source", None)
     if config.getini("ibutsu_source"):
         ibutsu_source = config.getini("ibutsu_source")
@@ -658,7 +723,8 @@ def pytest_configure(config):
                 ibutsu_server = ibutsu_server[:-1]
             if not ibutsu_server.endswith("/api"):
                 ibutsu_server += "/api"
-            ibutsu = IbutsuSender(ibutsu_server, ibutsu_source, extra_data=ibutsu_data)
+            ibutsu = IbutsuSender(ibutsu_server, ibutsu_source, extra_data=ibutsu_data,
+                                  token=ibutsu_token)
             ibutsu.frontend = ibutsu.health_api.get_health_info().frontend
         except MaxRetryError:
             print("Connection failure in health check - switching to archiver")
@@ -693,6 +759,13 @@ def pytest_unconfigure(config):
     if ibutsu_instance:
         del config._ibutsu
         config.pluginmanager.unregister(ibutsu_instance)
+        config.hook.pytest_ibutsu_before_shutdown(config=config, ibutsu=ibutsu_instance)
         ibutsu_instance.shutdown()
         if ibutsu_instance.run_id:
             ibutsu_instance.output_msg()
+
+
+def pytest_addhooks(pluginmanager):
+    import newhooks
+
+    pluginmanager.add_hookspecs(newhooks)
